#ifndef _GEORGE_CONSTANTS_H_
#define _GEORGE_CONSTANTS_H_

namespace george {

//
// Error values.
//
const int SOLVER_OK           = 0;
const int SETUP_FAILURE       = 1;
const int DIMENSION_MISMATCH  = 2;
const int CHOLMOD_ERROR       = 3;
const int MEMORY_ERROR        = 4;
const int USAGE_ERROR         = 5;

};

#endif
